@@ -20,15 +20,28 @@
 
 #include <string.h>
 
-#define G_PASTE_SEND_DBUS_SIGNAL_FULL(sig,data,error)               \
-    g_dbus_connection_emit_signal (priv->connection,                \
-                                   NULL, /* destination_bus_name */ \
-                                   G_PASTE_DAEMON_OBJECT_PATH,      \
-                                   G_PASTE_DAEMON_INTERFACE_NAME,   \
-                                   G_PASTE_DAEMON_SIG_##sig,        \
-                                   data,                            \
+#define G_PASTE_SEND_DBUS_SIGNAL_FULLER(interface, sig, data, error) \
+    g_dbus_connection_emit_signal (priv->connection,                 \
+                                   NULL, /* destination_bus_name */  \
+                                   G_PASTE_DAEMON_OBJECT_PATH,       \
+                                   interface,                        \
+                                   sig,                              \
+                                   data,                             \
                                    error)
 
+#define G_PASTE_SEND_DBUS_SIGNAL_FULL(sig,data,error) \
+    G_PASTE_SEND_DBUS_SIGNAL_FULLER (G_PASTE_DAEMON_INTERFACE_NAME, G_PASTE_DAEMON_SIG_##sig, data, error)
+
+#define G_PASTE_SEND_DBUS_PROPERTIES_CHANGED(property, value)      \
+    GVariantDict dict;                                             \
+    g_variant_dict_init (&dict, NULL);                             \
+    g_variant_dict_insert_value (&dict, property, value);          \
+    GVariant *data = g_variant_new ("(s@a{sv}@as)",                \
+                                    G_PASTE_DAEMON_INTERFACE_NAME, \
+                                    g_variant_dict_end (&dict),    \
+                                    g_variant_new_strv (NULL, 0)); \
+    G_PASTE_SEND_DBUS_SIGNAL_FULLER ("org.freedesktop.DBus.Properties", "PropertiesChanged", data, NULL)
+
 #define __NODATA     g_variant_new_tuple (NULL,  0)
 #define __DATA(data) g_variant_new_tuple (&data, 1)
 
@@ -200,7 +213,7 @@ g_paste_daemon_tracking (GPasteDaemon   *self,
     const GPasteDaemonPrivate *priv = _g_paste_daemon_get_instance_private (self);
     GVariant *variant = g_variant_new_boolean (tracking_state);
 
-    G_PASTE_SEND_DBUS_SIGNAL_WITH_DATA (TRACKING, variant);
+    G_PASTE_SEND_DBUS_PROPERTIES_CHANGED (G_PASTE_DAEMON_PROP_ACTIVE, variant);
 }
 
 /****************/
@@ -350,8 +363,9 @@ static void
 g_paste_daemon_private_delete (const GPasteDaemonPrivate *priv,
                                GVariant                  *parameters)
 {
-    g_paste_history_remove (priv->history,
-                            g_paste_daemon_get_dbus_uint64_parameter (parameters));
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
+
+    g_paste_history_remove_by_uuid (priv->history, uuid);
 }
 
 static void
@@ -408,19 +422,36 @@ static GVariant *
 g_paste_daemon_private_get_element (const GPasteDaemonPrivate *priv,
                                     GVariant                  *parameters,
                                     GPasteDBusError          **err)
+{
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
+    const GPasteItem *item = g_paste_history_get_by_uuid (priv->history, uuid);
+
+    G_PASTE_DBUS_ASSERT_FULL (item, "received no value for this uuid", NULL);
+
+    GVariant *variant = g_variant_new_string (g_paste_item_get_display_string (item));
+    return g_variant_new_tuple (&variant, 1);
+}
+
+static GVariant *
+g_paste_daemon_private_get_element_at_index (const GPasteDaemonPrivate *priv,
+                                             GVariant                  *parameters,
+                                             GPasteDBusError          **err)
 {
     GPasteHistory *history = priv->history;
     guint64 index = g_paste_daemon_get_dbus_uint64_parameter (parameters);
 
     G_PASTE_DBUS_ASSERT_FULL (index < g_paste_history_get_length (history), "invalid index received", NULL);
 
-    const gchar *value = g_paste_history_get_display_string (history, index);
+    const GPasteItem *item = g_paste_history_get (history, index);
 
-    G_PASTE_DBUS_ASSERT_FULL (value, "received no value for this index", NULL);
+    G_PASTE_DBUS_ASSERT_FULL (item, "received no value for this index", NULL);
 
-    GVariant *variant = g_variant_new_string (value);
+    GVariant *data[] = {
+        g_variant_new_string (g_paste_item_get_uuid (item)),
+        g_variant_new_string (g_paste_item_get_display_string (item))
+    };
 
-    return g_variant_new_tuple (&variant, 1);
+    return g_variant_new_tuple (data, 2);
 }
 
 static GVariant *
@@ -429,11 +460,9 @@ g_paste_daemon_private_get_element_kind (const GPasteDaemonPrivate *priv,
                                          GPasteDBusError          **err)
 {
     GPasteHistory *history = priv->history;
-    guint64 index = g_paste_daemon_get_dbus_uint64_parameter (parameters);
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
 
-    G_PASTE_DBUS_ASSERT_FULL (index < g_paste_history_get_length (history), "invalid index received", NULL);
-
-    const GPasteItem *item = g_paste_history_get (history, index);
+    const GPasteItem *item = g_paste_history_get_by_uuid (history, uuid);
 
     G_PASTE_DBUS_ASSERT_FULL (item, "received no item for this index", NULL);
 
@@ -449,26 +478,25 @@ g_paste_daemon_private_get_elements (const GPasteDaemonPrivate *priv,
 {
     GPasteHistory *history = priv->history;
     GVariantIter parameters_iter;
+    GVariantBuilder builder;
 
     g_variant_iter_init (&parameters_iter, parameters);
+    g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
 
     g_autoptr (GVariant) variant = g_variant_iter_next_value (&parameters_iter);
     guint64 len;
-    g_autofree guint64 *indexes = g_paste_util_get_dbus_at_result (variant, &len);
-    g_auto (GStrv) ans = g_new0 (gchar *, len + 1);
-    guint64 history_length = g_paste_history_get_length (history);
+    g_autofree const gchar **uuids = g_variant_get_strv (variant, &len);
 
     for (guint64 i = 0; i < len; ++i)
     {
-        G_PASTE_DBUS_ASSERT_FULL (indexes[i] < history_length, "invalid index received", NULL);
-        const gchar *value = g_paste_history_get_display_string (history, indexes[i]);
-        G_PASTE_DBUS_ASSERT_FULL (value, "received no value for this index", NULL);
-        ans[i] = g_strdup (value);
+        const GPasteItem *item = g_paste_history_get_by_uuid (history, uuids[i]);
+        G_PASTE_DBUS_ASSERT_FULL (item, "received no value for this index", NULL);
+        g_variant_builder_add (&builder, "(ss)", g_paste_item_get_uuid (item), g_paste_item_get_display_string (item));
     }
 
-    GVariant *answer = g_variant_new_strv ((const gchar * const *) ans, len);
+    GVariant *ans = g_variant_builder_end (&builder);
 
-    return g_variant_new_tuple (&answer, 1);
+    return g_variant_new_tuple (&ans, 1);
 }
 
 static GVariant *
@@ -476,13 +504,17 @@ g_paste_daemon_private_get_history (const GPasteDaemonPrivate *priv)
 {
     const GList *history = g_paste_history_get_history (priv->history);
     guint64 length = g_list_length ((GList *) history);
-    g_autofree const gchar **displayed_history = g_new (const gchar *, length + 1);
+    GVariantBuilder builder;
+
+    g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
 
     for (guint64 i = 0; i < length; ++i, history = g_list_next (history))
-        displayed_history[i] = g_paste_item_get_display_string (history->data);
-    displayed_history[length] = NULL;
+    {
+        const GPasteItem *item = history->data;
+        g_variant_builder_add (&builder, "(ss)", g_paste_item_get_uuid (item), g_paste_item_get_display_string (item));
+    }
 
-    GVariant *variant = g_variant_new_strv ((const gchar * const *) displayed_history, -1);
+    GVariant *variant = g_variant_builder_end (&builder);
 
     return g_variant_new_tuple (&variant, 1);
 }
@@ -523,17 +555,12 @@ g_paste_daemon_private_get_raw_element (const GPasteDaemonPrivate *priv,
                                         GVariant                  *parameters,
                                         GPasteDBusError          **err)
 {
-    GPasteHistory *history = priv->history;
-    guint64 index = g_paste_daemon_get_dbus_uint64_parameter (parameters);
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
+    const GPasteItem *item = g_paste_history_get_by_uuid (priv->history, uuid);
 
-    G_PASTE_DBUS_ASSERT_FULL (index < g_paste_history_get_length (history), "invalid index received", NULL);
-
-    const gchar *value = g_paste_history_get_value (priv->history, index);
-
-    G_PASTE_DBUS_ASSERT_FULL (value, "received no value for this index", NULL);
-
-    GVariant *variant = g_variant_new_string (value);
+    G_PASTE_DBUS_ASSERT_FULL (item, "received no value for this uuid", NULL);
 
+    GVariant *variant = g_variant_new_string (g_paste_item_get_value (item));
     return g_variant_new_tuple (&variant, 1);
 }
 
@@ -542,13 +569,17 @@ g_paste_daemon_private_get_raw_history (const GPasteDaemonPrivate *priv)
 {
     const GList *history = g_paste_history_get_history (priv->history);
     guint64 length = g_list_length ((GList *) history);
-    g_autofree const gchar **displayed_history = g_new (const gchar *, length + 1);
+    GVariantBuilder builder;
+
+    g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
 
     for (guint64 i = 0; i < length; ++i, history = g_list_next (history))
-        displayed_history[i] = g_paste_item_get_value (history->data);
-    displayed_history[length] = NULL;
+    {
+        const GPasteItem *item = history->data;
+        g_variant_builder_add (&builder, "(ss)", g_paste_item_get_uuid (item), g_paste_item_get_value (item));
+    }
 
-    GVariant *variant = g_variant_new_strv ((const gchar * const *) displayed_history, -1);
+    GVariant *variant = g_variant_builder_end (&builder);
 
     return g_variant_new_tuple (&variant, 1);
 }
@@ -580,27 +611,25 @@ g_paste_daemon_private_merge (const GPasteDaemonPrivate *priv,
 
     _variant_iter_read_strings_parameter (&parameters_iter, &decoration, &separator);
 
-    g_autoptr (GVariant) v_indexes = g_variant_iter_next_value (&parameters_iter);
+    g_autoptr (GVariant) v_uuids = g_variant_iter_next_value (&parameters_iter);
     guint64 length;
-    const guint64 *indexes = g_variant_get_fixed_array (v_indexes, &length, sizeof (guint64));
-
-    GPasteHistory *history = priv->history;
-    guint64 history_length = g_paste_history_get_length (history);
+    const GStrv uuids = (const GStrv) g_variant_get_strv (v_uuids, &length);
 
     G_PASTE_DBUS_ASSERT (length, "nothing to merge");
-    for (guint64 i = 0; i < length; ++i)
-    {
-        G_PASTE_DBUS_ASSERT (indexes[i] < history_length, "invalid index received");
-    }
 
+    GPasteHistory *history = priv->history;
     g_autoptr (GString) str = g_string_new (NULL);
 
     for (guint64 i = 0; i < length; ++i)
     {
+        const GPasteItem *item = g_paste_history_get_by_uuid (history, uuids[i]);
+
+        G_PASTE_DBUS_ASSERT (item, "no item matching this uuid");
+
         g_string_append_printf (str, "%s%s%s%s",
                                 (i) ? separator : "",
                                 decoration,
-                                g_paste_history_get_value (history, indexes[i]),
+                                g_paste_item_get_value (item),
                                 decoration);
     }
 
@@ -621,7 +650,6 @@ g_paste_daemon_track (GPasteDaemon *self,
     const GPasteDaemonPrivate *priv = _g_paste_daemon_get_instance_private (self);
 
     g_paste_settings_set_track_changes (priv->settings, tracking_state);
-    g_paste_daemon_tracking (self, tracking_state, NULL);
 }
 
 static void
@@ -667,13 +695,12 @@ g_paste_daemon_private_search (const GPasteDaemonPrivate *priv,
                                GVariant                  *parameters,
                                GPasteDBusError          **err)
 {
-    g_autofree gchar *name = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
-    g_autoptr (GArray) results = g_paste_history_search (priv->history, name);
+    g_autofree gchar *search = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
+    g_auto (GStrv) results = g_paste_history_search (priv->history, search);
 
     G_PASTE_DBUS_ASSERT_FULL (results, "Error while performing search", NULL);
 
-    GVariant *variant = g_variant_new_fixed_array (G_VARIANT_TYPE_UINT64, results->data, results->len, sizeof (guint64));
-
+    GVariant *variant = g_variant_new_strv ((const gchar * const *) results, -1);
     return g_variant_new_tuple (&variant, 1);
 }
 
@@ -682,12 +709,9 @@ g_paste_daemon_select (const GPasteDaemon *self,
                        GVariant           *parameters)
 {
     const GPasteDaemonPrivate *priv = _g_paste_daemon_get_instance_private (self);
-    guint64 position = g_paste_daemon_get_dbus_uint64_parameter (parameters);
-    GVariant *variant = g_variant_new_uint64 (position);
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
 
-    g_paste_history_select (priv->history, position);
-
-    G_PASTE_SEND_DBUS_SIGNAL_WITH_DATA (ITEM_SELECTED, variant);
+    g_paste_history_select (priv->history, uuid);
 }
 
 static void
@@ -702,11 +726,9 @@ g_paste_daemon_private_replace (const GPasteDaemonPrivate *priv,
     g_variant_iter_init (&parameters_iter, parameters);
 
     g_autoptr (GVariant) variant1 = g_variant_iter_next_value (&parameters_iter);
-    guint64 index = g_variant_get_uint64 (variant1);
+    const gchar *uuid = g_variant_get_string (variant1, NULL);
 
-    G_PASTE_DBUS_ASSERT (index < g_paste_history_get_length (history), "invalid index received");
-
-    const GPasteItem *item = g_paste_history_get (history, index);
+    const GPasteItem *item = g_paste_history_get_by_uuid (history, uuid);
 
     G_PASTE_DBUS_ASSERT (item, "received no item for this index");
     G_PASTE_DBUS_ASSERT (_G_PASTE_IS_TEXT_ITEM (item) && g_paste_str_equal (g_paste_item_get_kind (item), "Text"), "attempted to replace an item other than GPasteTextItem");
@@ -716,7 +738,7 @@ g_paste_daemon_private_replace (const GPasteDaemonPrivate *priv,
 
     G_PASTE_DBUS_ASSERT (contents, "no contents given");
 
-    g_paste_history_replace (priv->history, index, contents);
+    g_paste_history_replace (priv->history, uuid, contents);
 }
 
 static void
@@ -731,13 +753,10 @@ g_paste_daemon_private_set_password (const GPasteDaemonPrivate *priv,
     g_variant_iter_init (&parameters_iter, parameters);
 
     g_autoptr (GVariant) variant1 = g_variant_iter_next_value (&parameters_iter);
-    guint64 index = g_variant_get_uint64 (variant1);
+    const gchar *uuid = g_variant_get_string (variant1, NULL);
+    const GPasteItem *item = g_paste_history_get_by_uuid (history, uuid);
 
-    G_PASTE_DBUS_ASSERT (index < g_paste_history_get_length (history), "invalid index received");
-
-    const GPasteItem *item = g_paste_history_get (history, index);
-
-    G_PASTE_DBUS_ASSERT (item, "received no item for this index");
+    G_PASTE_DBUS_ASSERT (item, "received no item for this uuid");
     G_PASTE_DBUS_ASSERT (_G_PASTE_IS_TEXT_ITEM (item) && g_paste_str_equal (g_paste_item_get_kind (item), "Text"), "attempted to replace an item other than GPasteTextItem");
 
     g_autoptr (GVariant) variant2 = g_variant_iter_next_value (&parameters_iter);
@@ -746,7 +765,7 @@ g_paste_daemon_private_set_password (const GPasteDaemonPrivate *priv,
     G_PASTE_DBUS_ASSERT (name, "no password name given");
     G_PASTE_DBUS_ASSERT (!g_paste_history_get_password (priv->history, name), "a password with tat name already exists");
 
-    g_paste_history_set_password (priv->history, index, name);
+    g_paste_history_set_password (priv->history, uuid, name);
 }
 
 static void
@@ -784,19 +803,24 @@ g_paste_daemon_private_upload_finish (GObject      *source_object,
 /**
  * g_paste_daemon_upload:
  * @self: (transfer none): the #GPasteDaemon
- * @index: the index of the item to upload
+ * @uuid: the uuid of the item to upload
  *
  * Upload an item to a pastebin service
  */
 G_PASTE_VISIBLE void
 g_paste_daemon_upload (GPasteDaemon *self,
-                       guint64       index)
+                       const gchar  *uuid)
 {
     g_return_if_fail (_G_PASTE_IS_DAEMON (self));
 
     GPasteDaemonPrivate *priv = g_paste_daemon_get_instance_private (self);
+    const GPasteItem *item = (uuid) ? g_paste_history_get_by_uuid (priv->history, uuid) : g_paste_history_get (priv->history, 0);
+
+    if (!item)
+        return;
+
     GSubprocess *upload = g_subprocess_new (G_SUBPROCESS_FLAGS_STDIN_PIPE|G_SUBPROCESS_FLAGS_STDOUT_PIPE, NULL, "wgetpaste", NULL);
-    const gchar *value = g_paste_history_get_value (priv->history, index);
+    const gchar *value = g_paste_item_get_value (item);
 
     g_subprocess_communicate_utf8_async (upload,
                                          value,
@@ -809,7 +833,9 @@ static void
 _g_paste_daemon_upload (GPasteDaemon *self,
                         GVariant     *parameters)
 {
-    g_paste_daemon_upload(self, g_paste_daemon_get_dbus_uint64_parameter (parameters));
+    g_autofree gchar *uuid = g_paste_daemon_get_dbus_string_parameter (parameters, NULL);
+
+    g_paste_daemon_upload (self, uuid);
 }
 
 static void
@@ -871,6 +897,8 @@ g_paste_daemon_dbus_method_call (GDBusConnection       *connection     G_GNUC_UN
         g_paste_daemon_private_empty_history (priv, parameters);
     else if (g_paste_str_equal (method_name, G_PASTE_DAEMON_GET_ELEMENT))
         answer = g_paste_daemon_private_get_element (priv, parameters, &err);
+    else if (g_paste_str_equal (method_name, G_PASTE_DAEMON_GET_ELEMENT_AT_INDEX))
+        answer = g_paste_daemon_private_get_element_at_index (priv, parameters, &err);
     else if (g_paste_str_equal (method_name, G_PASTE_DAEMON_GET_ELEMENT_KIND))
         answer = g_paste_daemon_private_get_element_kind (priv, parameters, &err);
     else if (g_paste_str_equal (method_name, G_PASTE_DAEMON_GET_ELEMENTS))
